#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "cpgplot.h"

#define FORMAT_XSERVE 0
#define FORMAT_PS     1
#define FORMAT_GIF    2
#define FORMAT_PPM    3
#define FORMAT_VPS    4
#define FORMAT_VGIF   5
#define FORMAT_VPPM   6

#include "cora_fit.h"


static 
char * cut (double zahl, int n)
{
  int i;
  long ndigits;
  double vz, zround;
  char * cutthroat;
  static int nn = 0;
  static char cutt[2][17];

  if (n > 16 || n < 0)
    {
      fprintf(stderr, "* Warning *: cut: n (%d) must be positive < 16\n", n);
      return "0";
    }

  cutthroat = cutt[nn];
  nn = (nn == 0) ? 1 : 0;
      
  if (zahl != 0.0)
    {
      ndigits = (long)(log10(fabs(zahl))+.999);
      vz=zahl/fabs(zahl);
    }
  else
    {
      ndigits = 0.;
      vz=1.0;
    }

  if (n > 1) 
    zround=1.0*(zahl*pow(10.0,(n-ndigits)) + vz * 5) / pow (10.0,(n-ndigits));
  else 
    zround = zahl;

  for (i = 0; i < 17; ++i) 
    cutthroat[i] = '\0';

  sprintf(cutthroat, "%-16f", zround);
  for (i = 16; i >= 0; --i) 
    if (cutthroat[i] == ' ') cutthroat[i] = '\0';
  cutthroat[n] = '\0';

  return cutthroat;
}

/*  ------------------------ */

static 
float float_min (float * inArr, int size)
{
  int   i;
  float min = inArr[0];
  
  for (i = 0; i < size; ++i)
    min = (min < inArr[i]) ? min : inArr[i];

  return min;
}

static 
float float_max (float * inArr, int size)
{
  int   i;
  float max = inArr[0];
  
  for (i = 0; i < size; ++i)
    max = (max > inArr[i]) ? max : inArr[i];

  return max;
}

/*  ------------------------ */

static 
float * copy_to_float (Vector * inVec)
{
  int i;

  float * outVec = (float *) malloc (sizeof (float) * inVec->len);

  if (outVec == NULL)
    {
      fprintf (stderr, "* Warning *: copy_to_float: Out of memory\n");
      return NULL;
    }
  
  for (i = 0; i < inVec->len; ++i)
    outVec[i] = inVec->vec[i];

  return outVec;
}

/*  ------------------------ */

static char * check_name(char * name, char * ending)
{
  int size;
  int siz2;
  char * p;
  char myname[] = "cora";


  if (name == NULL)
    name = myname;

  size = strlen(name);
  siz2 = strlen(ending);

  if (size > siz2 && 0 == strcmp (&name[size-siz2], ending))
    {
      p = malloc (size+1);
      strcpy(p, name);
    }
  else
    {
      p = malloc (size+siz2+1);
      strcpy(p, name);
      strcat(p, ending);
    }
  return p;
}

/*  ------------------------ */

static 
float * rebin (Vector * inVec, int numBins, int flag)
{
  int i, j, k, delta, delta0;
  float ss;

  float * outVec = (float *) malloc (sizeof (float) * numBins);

  if (outVec == NULL)
    {
      fprintf (stderr, "*  Warning  *: rebin: Out of memory\n");
      return NULL;
    }
  if (numBins == 0)
    {
      fprintf (stderr, "*  Warning  *: rebin: numBins == 0\n");
      return NULL;
    }

  
  delta  = inVec->len / numBins;
  delta0 = delta;
  k      = inVec->len;
  j      = 0;

  while (1 == 1)
    {
      ss = 0.0;

      for (i = 0; i < delta; ++i)
	{
	  ss += inVec->vec[j*delta0 + i];
	}
      if (delta > 1 /* && flag == FALSE  ALWAYS conserve flux */) 
	ss /= delta;
      outVec[j] = ss;
      ++j;

      if (j == numBins)
	break;

      if ((j*delta0 + delta)  >=  inVec->len)
	delta = inVec->len - j*delta;
    }

  return outVec;
}

/*  ------------------------ */

int plotfit (char * fname       /* title             */,
	     char * psfile      /* dateiname         */,
	     int    fileformat  /* XSERVE/PS/GIF/PPM */
	     )
{
  /* global: ilog 0/1   1: Y-axis logarithmic */

/*  FILE *fd; */
  int    i, j, lline;
  int    numplot, numspec;
  int    flag_alloc;
  int    *index,flag=0;
  int    ldigs=5,edigs=0;
  double cs = 0.8;
  double signofA=1.0,yflip=0.;
  /* int    startgood, endgood; */
  float  xmin, xmax, tpos,t2pos;
  float  chi2, chitemp;
  int    chicount,ntest;
  float  tmp1, tmp2=0., tmp3=0.;
  float  spmax, spmin;
  float  maxext,minext;
  char   * str_open = NULL;
  char   theText[128],cpb[64],xunit[64];
  char   theText2[128];
  char   theText3[128];
  float  xAtext, yAtext, xpos, ypos, ml, moff;
  float  *slplot[MAXLINES];  /* single lines */

  Vector good_err, klam, ksig;

  float * logyplot, * log_sp_s, * log_sp_b,  * log_bg0_bin, 
        * log_sp_err, * log_sp_err1, * log_bg_instr;

  float * chi, * chil;

  float * xplot,  * yplot;           /* fitted                      */
  float * wav_s,  * sp_s, * sp_b;    /* original + background       */
  float * sp_err;                    /* errors                      */
  float * bg0_bin;                   /* original background         */
  float * bg_instr;		     /* instrumental background     */

  /* ----------  open plot ----------------
   */
  if (verbose)
   printf("Message      : Creating graph with fitted lines\n");

  if (psfile != NULL)
    str_open = (char *) malloc (strlen(psfile) + 32);
  else
    str_open = (char *) malloc (32);

  if (str_open == NULL)
    {
      fprintf (stderr, "* Warning *: plotfit: Out of memory\n");
      return (FALSE);
    }

  switch (fileformat)
    {
    case FORMAT_XSERVE:
      sprintf (str_open, "%s", "/XS");
      break;
    case FORMAT_PS:
      sprintf (psfile, "%s", check_name(psfile, ".ps"));
      sprintf (str_open, "%s/PS", psfile);
      break;
    case FORMAT_GIF:
      sprintf (str_open, "%s/GIF", check_name(psfile, ".gif"));
      break;
    case FORMAT_PPM:
      sprintf (str_open, "%s/PPM", check_name(psfile, ".ppm"));
      break;
    case FORMAT_VPS:
      sprintf (str_open, "%s/VPS", check_name(psfile, ".ps"));
      break;
    case FORMAT_VGIF:
      sprintf (str_open, "%s/VGIF", check_name(psfile, ".gif"));
      break;
    case FORMAT_VPPM:
      sprintf (str_open, "%s/VPPM", check_name(psfile, ".ppm"));
      break;

    }

  xplot = (float *) malloc (sizeof(float) * x.len);
  yplot = (float *) malloc (sizeof(float) * x.len);
  chi   = (float *) malloc (sizeof(float) * x.len);
  chil  = (float *) malloc (sizeof(float) * nlines);

  if (xplot == NULL || yplot == NULL || chi == NULL)
    {
      fprintf (stderr, "* Warning *: plotfit: Out of memory\n");
      free (str_open);
      return (FALSE);
    }

  vec_init (&good_err);
  vec_alloc(&good_err,  x.len);

  sel_vals (&good_err,  &err,  &good);

  if (nbin > 1)
    {
      wav_s    = rebin (&x,  x.len  / nbin, FALSE);
      sp_s     = rebin (&ncount,    ncount.len   / nbin, TRUE);
      sp_err   = rebin (&good_err,  good_err.len  / nbin, TRUE);
      bg_instr = rebin (&bbb, bbb.len / nbin, TRUE);
      sp_b     = rebin (&bbb, bbb.len / nbin, TRUE);
      bg0_bin  = rebin (&bb0, bb0.len / nbin, TRUE);
      numspec  = count / nbin;
      flag_alloc = TRUE;
    }
  else
    {
      wav_s    = copy_to_float(&x);
      sp_s     = copy_to_float(&ncount);
      sp_err   = copy_to_float(&good_err);
      sp_b     = copy_to_float(&bbb);
      bg_instr = copy_to_float(&bbb);
      bg0_bin  = copy_to_float(&bg0);
      numspec  = count;
      flag_alloc = TRUE;
    }

  dlam = fabs(dlam);

  /* define plot limits
   */
  tmp1 = min(&llambda) - dlam; tmp2 = min(&x);
  xmin = (tmp1 > tmp2) ? tmp1 : tmp2;
  tmp1 = max(&llambda) + dlam; tmp2 = max(&x);
  xmax = (tmp1 < tmp2) ? tmp1 : tmp2;
    
  /* sum gauss profiles in matrix yc for all lines, but only
   * within plot range
   */
  numplot = 0;

  for(i=0; i < nlines; ++i) 
    {
      slplot[i] = malloc(x.len * sizeof(float));

      if (slplot[i] == NULL)
        {
          fprintf (stderr, "*  Warning  *: Out of memory\n"); 
          exit (8);
        }

      /* initialize slplot 
       */
      for (j=0; j < count; ++j) 
	slplot[i][j]=0.0;
    }

  for (j = 0; j < count; ++j)
    {
      chi[j] = 0.0;
      for (i = 0; i < nlines; ++i)
	{
	 chil[i] = yc[i][j] * ampout.vec[i];
	 chi[j] += chil[i];
	}
      chi[j] += bbb.vec[j];
      if (x.vec[j] >= xmin && x.vec[j] <= xmax)
	{
	  yplot[numplot] = chi[j];
	  for (i=0; i<nlines; ++i) 
	    slplot[i][numplot] += chil[i]+bbb.vec[j];
	  xplot[numplot] = x.vec[j];
	  ++numplot;
	}
    }
  chi2 = 0.0; chicount = 0;
  for (i = 0; i < count; ++i)
    {
      if (good_err.vec[i] > 0.0)
	{
	  ++chicount;
	  chitemp = (chi[i]-ncount.vec[i]);
	  chi2 += (chitemp * chitemp)/(good_err.vec[i] * good_err.vec[i]);
	}
    }
  if (chicount>0) chi2 /= chicount; else chi2 = -1.;
 

  if (numplot == 0)
    {
      fprintf (stderr, "*  Warning  *: plotfit: Empty plot range (%f - %f)\n",
	       xmin, xmax);
      fprintf (stderr, 
	       "*  Warning  *: plotfit: (possible reason: dsig too small)\n");
      free (str_open);
      vec_free (&good_err);
      if (flag_alloc == TRUE)
	{
	  free (wav_s); free (sp_s); free (sp_err);
	  free (sp_b); free (bg0_bin); free (bg_instr);
	}
      return (FALSE);
    }

  if(icount==0)
   {
     for(i=0;i<numspec;++i)
      {
	sp_s[i]    = sp_s[i]/binsize;
	bg_instr[i]= (sp_b[i]-sbg)/binsize;
	sp_b[i]    = sp_b[i]/binsize;
	bg0_bin[i] = bg0_bin[i]/binsize;
	sp_err[i]  = sp_err[i]/binsize;
      }
     for(i=0;i<numplot;++i)
	{
	 yplot[i] /= binsize;
	 for (j=0;j<nlines;++j) slplot[j][i] /= binsize;
	}
   }
  else
   {
     for(i=0;i<numspec;++i)
      {
	sp_s[i]    = sp_s[i]*nbin;
	bg_instr[i]= (sp_b[i]-sbg)*nbin;
	sp_b[i]    = sp_b[i]*nbin;
	bg0_bin[i] = bg0_bin[i]*nbin;
	sp_err[i]  = sp_err[i]*nbin;
      }
     for(i=0;i<numplot;++i)
	{
	   yplot[i] *= nbin;
	   for (j=0;j<nlines;++j) slplot[j][i] *= nbin;
	}
   }

/*     subtract instrumental background for plot 
 for (i=0;i<numspec;++i) sp_s[i] = sp_s[i]-bg_instr[i]; */

  tmp1 = yplot[0]; tmp2 = yplot[0];
  for (i = 0; i < numplot; ++i)
    {
      tmp1 = (tmp1 > yplot[i]) ? tmp1 : yplot[i];
      tmp2 = (tmp2 < yplot[i]) ? tmp2 : yplot[i];
    }
  spmax = float_min(sp_s,numspec); spmin = float_max(sp_s,numspec);
  for (i = 0; i < numspec; ++i)
    {
      if (wav_s[i] >= xmin && wav_s[i] <= xmax)
        {
          spmax = (spmax > sp_s[i]) ? spmax : sp_s[i];
          spmin = (spmin < sp_s[i]) ? spmin : sp_s[i];

/*	  spmax = (spmax > bg_instr[i]) ? spmax : bg_instr[i];
	  spmin = (spmin < bg_instr[i]) ? spmin : bg_instr[i];  */
        }
    }
  spmax = (spmax > tmp1) ? spmax : tmp1; 
  spmin = (spmin < tmp2) ? spmin : tmp2; 

  maxext = (spmax-spmin)*0.3;
  minext = (spmax-spmin)*0.3;
  if (ilog==1 && spmin-minext<=0.) minext = 0.0;
  spmax += maxext;
  spmin -= minext;
  /* if (spmin < 0.0) */ spmin = 0.0;

  vec_init (&klam);
  vec_init (&ksig);
  vec_alloc(&klam, nlines);
  vec_alloc(&ksig, nlines);

if (ikev == 1)
 {
   tmp1=xmin;
   xmin=12.398544/xmax;
   xmax=12.398544/tmp1;
   dlam=0;
  sprintf(xunit,"E / keV");

   for (i=0;i<nlines;++i)
     {
	klam.vec[i] = 12.398544 / llambda.vec[i];
	ksig.vec[i] = 12.398544 / lsigma.vec[i];
     }
   for (i=0;i<numspec;++i)
	{
	  wav_s[i]      = 12.398544 / wav_s[i];
	  j = numspec-i-1;
	   tmp1 = sp_s[i];
	  sp_s[i]       = sp_s[j];
	  sp_s[j]	= tmp1;
	   tmp1=sp_err[i];
	  sp_err[i]     = sp_err[j];
	  sp_err[j]	= tmp1;
	   tmp1=sp_b[i];
	  sp_b[i]       = sp_b[j];
	  sp_b[j]	= tmp1;
	   tmp1=bg_instr[i];
	  bg_instr[i]   = bg_instr[j];
	  bg_instr[j]	= tmp1;
	}
  for (i=0;i<numplot;++i)
	{
	  xplot[i]	= 12.398544 / xplot[i];
	  j = numplot-i-1;
	   tmp1=yplot[i];
	  yplot[i]=yplot[j];
	  yplot[j]=tmp1;
	}
 }
else {
  for (i=0;i<nlines;++i) klam.vec[i] = llambda.vec[i];
  sprintf(xunit,"\\gl\\fn / \\A");}

  /* plot
   */

  if(cpgopen(str_open) <= 0)
    {
      fprintf (stderr, "*  Warning  *: plotfit: Cannot open plot device\n");
      free (str_open);
      return (FALSE);
    }

  ntest=abs(floor(log10(binsize*nbin)))+2;
  sprintf(cpb,"Counts/%s \\A bin",cut(binsize*nbin,ntest));
  if (exposure != 1.) sprintf(cpb,"%s/%sksec",cpb,cut(exposure/1000.,5));
  if (ilog == 1)
    {
      logyplot     = (float *) malloc (sizeof(float) * numplot);
      log_sp_s     = (float *) malloc (sizeof(float) * numspec);
      log_sp_b     = (float *) malloc (sizeof(float) * numspec);
      log_bg_instr = (float *) malloc (sizeof(float) * numspec);
      log_bg0_bin  = (float *) malloc (sizeof(float) * numspec);
      log_sp_err   = (float *) malloc (sizeof(float) * numspec);
      log_sp_err1  = (float *) malloc (sizeof(float) * numspec);

      if (logyplot == NULL    || log_sp_s == NULL    || log_sp_b == NULL ||
	  log_sp_err == NULL  || log_sp_err1 == NULL || 
	  log_bg0_bin == NULL || log_bg_instr == NULL)
	{
	  fprintf (stderr, "*  Warning  *: plotfit: Out of memory\n");
	  cpgclos();
	  return (FALSE);
	}

      for (i = 0; i < numspec; ++i)
	{
	  log_sp_s[i]     = log10(sp_s[i]);
	  log_sp_b[i]     = log10(sp_b[i]);
	  log_bg_instr[i] = log10(sp_b[i]-sbg);
	  log_bg0_bin[i]  = log10(log_bg0_bin[i]);
	  log_sp_err[i]  = log10(sp_s[i] + sp_err[i]);
	  log_sp_err1[i] = log10(sp_s[i] - sp_err[i]);
	}
      for (i = 0; i < numplot; ++i)
	{
	  logyplot[i] = log10(yplot[i]);
	}

      cpgsch(1.3);
      if (spmin == 0 ) spmin = float_min(bg_instr,numspec);
      cpgenv(xmin, xmax, log10(spmin), log10(spmax), 0, 20);
      if (icount==0) cpglab(xunit, "Counts / \\A", fittitle);
         else cpglab("\\gl\\fn / \\A", cpb, fittitle);
      cpgqlw(&i);      
      if (fit_only == 0)
       {
	cpgslw(4*i);
	cpgbin (numspec, wav_s, log_sp_s, 1);
       }
      if (spec_only == 0)
       {
	cpgslw(2*i);      
	cpgline (numplot, xplot, logyplot);
	cpgsls(2);
	cpgline (numspec, wav_s, log_sp_b);
	cpgsls(3);
	cpgline (numspec, wav_s, log_bg_instr);
/*	for (i=0;i<nlines;++i) cpgline (numplot, xplot, slplot[i]); */
       }
      cpgsls(1);
      if (ierrplot > 0)
	cpgerry (numspec, wav_s, log_sp_err, log_sp_err1, 1.0);
    }
  else
    {
      cpgsch(1.4);
      cpgqlw(&i);
/*	i=.01*i;   reduce line thickness */
      cpgslw(3*i);
      cpgenv(xmin, xmax, (spmin), (spmax), 0, 0);
      if (icount==0) cpglab(xunit, "Counts / \\A", "");
         else cpglab(xunit, cpb, "");
/*  print a subtitle (need to adapt scales) */
/*  cpgtext(13.45,spmax+.02*(spmax-spmin),"fit consistent with (f+i)/r=0.8"); */
  cpgsch(1.8);
  cpglab("", "", fittitle);
  cpgsch(1.4);
      if (fit_only == 0)
       {
	cpgslw(4*i);
	cpgbin (numspec, wav_s, sp_s, 1);
       }
      if (spec_only == 0)
       {
	cpgslw(2*i);
	cpgline (numplot, xplot, yplot);
	cpgsls(2);
	cpgline (numspec, wav_s, sp_b);
	cpgsls(3);
	if (nlines > 1) for (i=0;i<nlines;++i) cpgline (numplot, xplot, slplot[i]);
/* fd = fopen ("fit.spec","w");  write fit into a file
for (i=0;i<numplot;++i)
        fprintf(fd,"%8.3f %4.2f\n",xplot[i],yplot[i]);
     fclose(fd);
 */
       }
      cpgsls(1);
      if (ierrplot > 0)
	cpgerrb (6, numspec, wav_s, sp_s, sp_err, 0.0);
      if (spec_only == 0 && fit_only == 0)
	{
	 cpgsls(3);
	 cpgline (numspec, wav_s, bg_instr);
	}
    }

/*      tpos = (xmax-xmin)*0.23;
      t2pos= 0.09*(spmax-spmin);

cpgsch(2.0);
cpgscf(2);
cpgtext(xmax-tpos, spmax-t2pos, fittitle);
cpgsch(1.4);
 */

  if (ilab > 0)  /* labels */
    {
      cpgscf(2);
      ml = fabs(xplot[1] - xplot[0]);
      moff=.1/max(&ampout);

      cpgsch(1.0/(1.4*cs));

      index = (int *) malloc (sizeof(int) * nlines);
      tmp1=min(&klam);
      for(i = 0; i < nlines; ++i) if(fabs(klam.vec[i]- tmp1)<.01) index[0]=i;
      for(lline = 1; lline < nlines; ++lline)
	{
        tmp1=max(&klam);
	for(i=0;i<nlines;++i)
	 {
	   flag=0;
	   for(j=0; j<lline; ++j) 
	      if (i==index[j]) flag=1;
	   if(flag==0)
	      if(klam.vec[i]-tmp1<0.001)
	       {
		tmp1=klam.vec[i];
		index[lline]=i;
	       }
	  }
	}

      for(lline = 0; lline < nlines; ++lline)
        {
         i = index[lline];
	 signofA = (abs(ampout.vec[i]) > 0.0) ? ampout.vec[i]/fabs(ampout.vec[i]) : 1.;

	  /* ------- AMPLITUDE ------------ */
	  if (eramp.vec[i] > 0)
	    sprintf (theText, "A=%7s\\(2233)%5s", cut(ampout.vec[i], 7), 
		     cut (eramp.vec[i], 7));
	  else
	    sprintf (theText, "A=%7s", cut(ampout.vec[i], 7));
	  sprintf(theText,  "%scts",theText);

	  xAtext = klam.vec[i];
	  yAtext = slplot[i][0];
	  for (j = 0; j < numplot; ++j)  /* find the line maximum */
	    {
	      if (xplot[j] >= (klam.vec[i]-ml) && 
		  xplot[j] <= (klam.vec[i]+ml))
		yAtext = (signofA*yAtext > signofA*slplot[i][j]) ? yAtext : slplot[i][j];
	    } 

/*	  yAtext = (1.0+signofA*moff*(spmax-spmin)/nbin) * yAtext; */
	  if (ilog == 1)
	   cpgtext(xAtext, log10(yAtext), theText);
	  else
	   cpgtext(xAtext, yAtext, theText);


	  /* ------ SIGMA ------------------ */

	  
	  if (ersig.vec[i] > 1.0e-4)
	   {
	    sprintf(theText,  "%5s\\(2233)%6s", 
		     cut(lsigma.vec[i],5), cut(ersig.vec[i],6));
	   }
	  else
	   {
	    sprintf(theText,  "%5s", cut(lsigma.vec[i],5));
	   }
	  if (sigunit == 0)  
	   {
	    if (ikev == 1) tmp1 = -1; else tmp1 = 1;
	    xpos = llambda.vec[i] + tmp1*1.3 * lsigma.vec[i];
	    sprintf(theText,  "%s\\A",theText);
	    if (ikev == 1) xpos = 12.398544 / xpos;
	   }
	  else
	   {
	    if (ikev == 1) tmp1 = -1; else tmp1 = 1;
	    xpos = llambda.vec[i] + tmp1* 1.3 *(llambda.vec[i]*lsigma.vec[i]/C_LIGHT);
	    sprintf(theText,  "%skm/s",theText);
            if (ikev == 1) xpos = 12.398544 / xpos;
	   }

	  if (ilog == 1)
	   ypos = log10(.9*float_min(sp_b, numspec) + yAtext/6.);
	  else
	   ypos = (yAtext + float_min(sp_b, numspec))/2.;
	  cpgtext(xpos, ypos, theText);

	  /* ------ LAMBDA ----------------- */
	  
	if (fit_lock[lline] == 0)
	 {
	  xpos = klam.vec[i] - 0.85 * (xmax - xmin) * binsize;

	  if (lline > 1  && xpos < tmp3 && yflip != 1.5) yflip=1.5; else
          if (lline > 0  && xpos < tmp2 && yflip==0) yflip=.7; else yflip=0.;

	  tmp3 = tmp2; /* last end position */

	  ldigs = ceil(log(llambda.vec[i]))+3;  /* number of digits */
	  if (erlam.vec[i] > 0.) edigs = ceil(abs(log(erlam.vec[i]))); else edigs=0.;
	  tmp2 = xpos+(xmax-xmin)/0.54*0.85*(ldigs+edigs+3)*binsize;
			/* current end position of label*/

	  if (erlam.vec[i] > 1.0e-6)
	    sprintf(theText, "%7s\\(2233)%6s", cut(llambda.vec[i],ldigs), 
		    cut(erlam.vec[i],edigs));
	  else
	    {sprintf(theText, "%7s", cut(llambda.vec[i],ldigs));edigs=0.;}
	  sprintf(theText,  "%s\\A",theText);

	  if (ilog == 1)
	    ypos = log10(.9*spmin+(spmax-spmin)*.01 + log10(spmin)/100.0 * (spmax-spmin) * yflip);
	  else
	    ypos = (spmax-spmin) * 0.05 * yflip+spmin+.01*(spmax-spmin);
          cpgtext(xpos, ypos, theText);
	 }
	  
	}

      /* ----------- other interesting stuff ------- */

      /* RIGHT corner
       */

      if (fname[strlen(fname)-1] == '\n')
	fname[strlen(fname)-1] = '\0';

/*      if (dt>0.) sprintf(theText,  "t=%4s ksec", cut(dt, 4));  */
      sprintf(theText2, "BG\\dsource\\u=%4scts/\\A", 
	      cut((sbg/(wave.vec[(int)good.vec[count-1]] - 
			wave.vec[(int)good.vec[0]])*(count-1) ),4));
      sprintf(theText3, "\\gx\\fn\\u2\\d/N=%5s", cut(chi2, 5));

      tpos = (xmax-xmin)*0.23;
      t2pos= 0.05*(spmax-spmin);

      if (ilog == 1)
	{
	  spmax *= .95;
	  cpgtext(xmax-tpos, log10(spmax*.97), fname);
/*	  cpgtext(xmax-tpos, log10(spmax*.9), theText); */
	  if (chi2 >= 0.0)
	     cpgtext(xmax-tpos, log10(spmax*.9), theText3);
	  if (sbg > 0.0)
	    cpgtext(xmax-tpos, log10(spmax*.83), theText2);
	}
      else
	{
	  cpgtext(xmax-tpos, spmax-t2pos, fname);
/*	  cpgtext(xmax-tpos, spmax-t2pos*1.9, theText); */
	  if (chi2 >= 0.0)
	     cpgtext(xmax-tpos, spmax-t2pos*1.9, theText3);
	  if (sbg > 0.0)
	    {
	     if (chi2 < 0.0) cpgtext(xmax-tpos, spmax-t2pos*1.9, theText2);
	     else cpgtext(xmax-tpos, spmax-t2pos*3.0, theText2);
	    }
	}	  

      /* LEFT corner
       */

      if (ilog == 1)
	{
	  cpgtext(xmin+dlam/10., log10(spmax*0.97), "Iterated variables:");
	  tpos = 0.9;
	  if (ilamit > 0)
	    {
	      if (idlam > 0 && nlines>1)
		sprintf(theText, "\\gl\\fn\\drestricted");
	      else
		sprintf(theText, "\\gl");
	      cpgtext(xmin+dlam/10., log10(spmax*tpos), theText);
	      tpos -= 0.07;
	    }
	  if (isigit > 0)
	    {
	      sprintf(theText, "\\gs");
	      cpgtext(xmin+dlam/10., log10(spmax*tpos), theText);
	      tpos -= 0.07;
	    }
	  if (sbgit > 0)
	    {
	      sprintf(theText, "BG\\dsource");
	      cpgtext(xmin+dlam/10., log10(spmax*tpos), theText);
	    }
	}
      else
	{
	  cpgtext(xmin+dlam/10., spmax-t2pos, "Iterated variables:");
	  tpos   = t2pos;
	  t2pos *= 1.9;
	  tpos   = t2pos/tpos;

	  if (ilamit > 0)
	    {
	      if (idlam > 0 && nlines>1)
		sprintf(theText, "\\gl\\fn\\drestricted");
	      else
		sprintf(theText, "\\gl");
	      cpgtext(xmin+dlam/10., spmax-t2pos, theText);
	      t2pos *= 2.-1./tpos;
	      tpos   = 2.-1./tpos;
	    }
	  if (isigit > 0)
	    {
	      sprintf(theText, "\\gs");
	      cpgtext(xmin+dlam/10., spmax-t2pos, theText);
	      t2pos *= 2.-1./tpos;
	    }
	  if (sbgit > 0)
	    {
	      sprintf(theText, "BG\\dsource");
	      cpgtext(xmin+dlam/10., spmax-t2pos, theText);
	    }
	}
	

    }

  cpgclos();

  if (str_open != NULL) 
    free (str_open);
  vec_free (&good_err);
  if (flag_alloc == TRUE)
    {
      free (wav_s); free (sp_s); free (sp_err); free (sp_b);
      free (bg0_bin); free (bg_instr);
    }

  return (TRUE);
}
